package com.cloudhopper.jni;

/*
 * #%L
 * ch-jni-loader
 * %%
 * Copyright (C) 2012 - 2013 Cloudhopper by Twitter
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * A collection of libraries embedded in a JAR. After construction, call unpack()
 * to initialize.
 *
 * @author garth
 */
public class EmbeddedLibraries extends Debug {

    private static final String FILE_SEPARATOR = System.getProperty("file.separator");
    private static final int TEMP_DIR_ATTEMPTS = 10000;

    public EmbeddedLibraries(Class clazz) {
	this.clazz = clazz;
	this.os = OS.getOS();
	this.arch = Arch.getArch();
    }

    public EmbeddedLibraries(Class clazz, OS os, Arch arch) {
	this.clazz = clazz;
	this.os = os;
	this.arch = arch;
    }

    private final Class clazz;
    private final OS os;
    private final Arch arch;

    private File directory = null;
    private File[] libraryCatalog = null;

    public Class getInitializingClass() { return this.clazz; }
    public OS getOS() { return this.os; }
    public Arch getArch() { return this.arch; }

    public File getDirectory() { 
	if (directory == null) throw new IllegalStateException("Libraries have not been unpacked. Call unpack() first.");
	else return directory;
    }

    public File[] getLibraryCatalog() {
	if (libraryCatalog == null) throw new IllegalStateException("Libraries have not been unpacked. Call unpack() first.");
	else return libraryCatalog;
    }

    public String getPath() {
	return getPath("native", os, arch);
    }

    public String toString() {
	StringBuffer o = new StringBuffer();
	o.append("{");
	o.append("initializingClass=").append(clazz).append(", ");
	o.append("os=").append(os).append(", ");
	o.append("arch=").append(arch).append(", ");
	o.append("directory=").append(directory).append(", ");
	o.append("libraryCatalog=").append("[");
	for (File l:libraryCatalog) {
	    o.append(l.getPath()).append(",");
	}
	o.append("]}");
	return o.toString();
    }

    public static String getPath(String prefix, OS os, Arch arch) {
	return FILE_SEPARATOR+prefix+FILE_SEPARATOR+os.name()+FILE_SEPARATOR+arch.name()+FILE_SEPARATOR;
    }

    synchronized public void unpack() throws IOException {
	String path = getPath();
	if (directory == null) {
	    directory = createTempDir();
	    directory.deleteOnExit();
	    String[] resources = listResources(path.substring(1));
	    if (resources.length > 0) {
		int numFiles = 0;
		libraryCatalog = new File[resources.length];
		for (String resource : resources) {
		    debug(resource);
		    File tmpFile = new File(directory.getPath()+FILE_SEPARATOR+resource);
		    loadInto(path, resource, tmpFile);
		    tmpFile.deleteOnExit();
		    libraryCatalog[numFiles] = tmpFile;
		    numFiles++;
		}
	    }
	} else {
	    throw new IllegalStateException("Already unpacked into "+directory.getPath());
	}
    }

    private void loadInto(String path, String lib, File file) throws IOException {
	URL nativeUrl = clazz.getResource(path+lib);
	
	final InputStream in = nativeUrl.openStream();
	final OutputStream out = new BufferedOutputStream(new FileOutputStream(file));
	
	int len = 0;
	byte[] buffer = new byte[8192];
	while ((len = in.read(buffer)) > -1) out.write(buffer, 0, len);
	out.close();
	in.close();
	
	//System.load(libfile.getAbsolutePath());
	//libfile.delete();
    }

    /**
     * List the resources in the ClassLoader of the given Class that exist in the given relative directory path.
     * ADAPTED FROM: http://www.uofr.net/~greg/java/get-resource-listing.html
     *  - non-recursive
     *  - ignores included directories
     *  - omits .
     */
    private String[] listResources(String path) throws IOException {
	debug("Listing resources for "+clazz+" at "+path);

	Package pkg = clazz.getPackage();
	debug("Package for class "+clazz+" is "+pkg.getName());

	ClassLoader classLoader = clazz.getClassLoader();
	debug("ClassLoader is "+classLoader);

	URL dirURL = null;
	Set<String> result = new HashSet<String>(); //avoid duplicates in case it is a subdirectory

	Enumeration<URL> urls = classLoader.getResources(path);

	while (urls.hasMoreElements()) {
	    URL u = urls.nextElement();
	    debug("Found resource at "+u+" for "+path);
	    // if (u.getPath().contains("com/cloudhopper/kyotocabinet")) {
	    // 	dirURL = u;
	    // 	break;
	    // }

	    if (u != null && u.getProtocol().equals("file")) {
		try {
		    String[] tmpFiles = new File(u.toURI()).list();
		    for (String tmpFile:tmpFiles) {
			result.add(tmpFile);
		    }
		} catch (Exception ignore) {}
	    } 
	    
	    // if (u == null) {
	    //     String me = clazz.getName().replace(".", "/")+".class";
	    //     u = clazz.getClassLoader().getResource(me);
	    // }
	    
	    if (u.getProtocol().equals("jar")) {
		String jarPath = u.getPath().substring(5, u.getPath().indexOf("!")); //strip out only the JAR file
		JarFile jar = new JarFile(URLDecoder.decode(jarPath, "UTF-8"));
		Enumeration<JarEntry> entries = jar.entries(); //gives ALL entries in jar
		while(entries.hasMoreElements()) {
		    String name = entries.nextElement().getName();
		    if (name.startsWith(path)) { //filter according to the path
			String entry = name.substring(path.length());
			int checkSubdir = entry.indexOf("/");
			if (checkSubdir >= 0) {
			    //dont add subdirs
			} else if (entry.trim().equals("")) {
			    //dont add .
			} else {
			    result.add(entry);
			}
		    }
		}
	    }
	}
	return result.toArray(new String[0]);
	//return new String[0];
    }

    /**
     * Creates a temporary directory with the base location equal to the 'java.io.tmpdir'
     * varibal. Prevents the File.createTempFile delete()/mkdirs() race condition. Attempts
     * to make the directory a configurable number of times before throwing an 
     * IllegalStateException.
     */
    private File createTempDir() {
	File baseDir = new File(System.getProperty("java.io.tmpdir"));
	String baseName = System.currentTimeMillis() + "-";
	
	for (int counter = 0; counter < TEMP_DIR_ATTEMPTS; counter++) {
	    File tempDir = new File(baseDir, baseName + counter);
	    if (tempDir.mkdir()) {
		return tempDir;
	    }
	}
	throw new IllegalStateException("Failed to create directory within "
					+ TEMP_DIR_ATTEMPTS + " attempts (tried "
					+ baseName + "0 to " + baseName + (TEMP_DIR_ATTEMPTS - 1) + ')');
    }

}
