package com.cloudhopper.jni;

/*
 * #%L
 * ch-jni-loader
 * %%
 * Copyright (C) 2012 - 2013 Cloudhopper by Twitter
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import java.util.*;
import java.util.jar.*;
import java.io.*;
import java.net.*;
import java.lang.reflect.Field;

/**
 * Some approaches for loading os/architecture-specific native libraries, both from the
 * library path and embedded in the jar. Warning that this contains some dangerous
 * routines that may not work in security managers or may not be compatible with all
 * JVM versions.
 *
 * @author garth
 */
public class NativeLoader extends Debug {
    
    private static NativeLoader instance;
    
    static {
	instance = new NativeLoader();
    }
    
    private Set<String> loadedLibraries;
    private Map<ClassLoader,EmbeddedLibraries> classLoaderPaths;
    
    private NativeLoader() {
	this.loadedLibraries = new HashSet<String>();
	this.classLoaderPaths = new HashMap<ClassLoader,EmbeddedLibraries>();
    }
    
    /**
     * Load a native library from the ClassLoader of the given Class.
     * @param clazz The Class to use to identify the ClassLoader where the native library resides.
     * @param libName The name of the native library to load.
     */
    synchronized public static void load(Class clazz, String libName) throws SecurityException, UnsatisfiedLinkError, IOException, IllegalAccessException {
	instance.doLoad(clazz, libName);
    }
    
    private void doLoad(Class clazz, String libName) throws SecurityException, UnsatisfiedLinkError, IOException, IllegalAccessException {
	if (loadedLibraries.contains(libName)) {
	    debug(libName+" is already loaded in loadedLibraries map");
	    return;
	} else {
	    loadLibrary(clazz, libName);
	    loadedLibraries.add(libName);
	}
    }

    private void loadLibrary(Class clazz, String libname) throws SecurityException, UnsatisfiedLinkError, IOException, IllegalAccessException {
	debug("System.mapLibraryName for libname = "+System.mapLibraryName(libname));

	// Try preinstalled libs
	try {
	    System.loadLibrary(libname);
	    return;
	} catch (UnsatisfiedLinkError le) {
	    debug("Unable to load from preinstalled libraries", le);
	}

	ClassLoader classLoader = clazz.getClassLoader();
	EmbeddedLibraries embeddedLibs = classLoaderPaths.get(classLoader);
	if (embeddedLibs == null) {
	    embeddedLibs = new EmbeddedLibraries(clazz);
	    embeddedLibs.unpack();
	    debug("Storing "+embeddedLibs+" in loadedLibraries map with ClassLoader key "+classLoader);
	    classLoaderPaths.put(classLoader, embeddedLibs);
	} else {
	    debug(classLoader+" was found in loadedLibraries map.");
	}

	String initialLibraryPath = System.getProperty("java.library.path");
	try {
	    setJavaLibraryPath(initialLibraryPath+":"+embeddedLibs.getDirectory());
	    System.loadLibrary(libname);
	} finally {
	    // set java.library.path back to the original value
	    setJavaLibraryPath(initialLibraryPath);
	}
    }

    /**
     * Set the java.library.path System variable.
     * ADAPTED FROM: http://blog.cedarsoft.com/2010/11/setting-java-library-path-programmatically/
     */
    private void setJavaLibraryPath(String path) throws SecurityException, IllegalAccessException {
	try {
	    System.setProperty("java.library.path", path);
	    Field fieldSysPath = ClassLoader.class.getDeclaredField("sys_paths");
	    fieldSysPath.setAccessible(true);
	    fieldSysPath.set(null, null);
	    debug("New java.library.path = "+System.getProperty("java.library.path"));
	} catch (NoSuchFieldException e) {
	    debug("This shouldn't happen. sys_paths is always present in ClassLoader.", e);
	}
    }

}
